package kr.co.vividnext.sodalive.content.main

import kr.co.vividnext.sodalive.content.AudioContentRepository
import kr.co.vividnext.sodalive.content.main.banner.AudioContentBannerType
import kr.co.vividnext.sodalive.content.main.banner.GetAudioContentBannerResponse
import kr.co.vividnext.sodalive.content.main.curation.GetAudioContentCurationResponse
import kr.co.vividnext.sodalive.content.theme.AudioContentThemeQueryRepository
import kr.co.vividnext.sodalive.event.EventItem
import kr.co.vividnext.sodalive.member.Member
import kr.co.vividnext.sodalive.member.block.BlockMemberRepository
import org.springframework.beans.factory.annotation.Value
import org.springframework.cache.annotation.Cacheable
import org.springframework.data.domain.Pageable
import org.springframework.stereotype.Service
import org.springframework.transaction.annotation.Transactional

@Service
class AudioContentMainService(
    private val repository: AudioContentRepository,
    private val blockMemberRepository: BlockMemberRepository,
    private val audioContentThemeRepository: AudioContentThemeQueryRepository,

    @Value("\${cloud.aws.cloud-front.host}")
    private val imageHost: String
) {
    @Transactional(readOnly = true)
    @Cacheable(cacheNames = ["default"], key = "'themeList:' + ':' + #isAdult")
    fun getThemeList(isAdult: Boolean): List<String> {
        return audioContentThemeRepository.getActiveThemeOfContent(isAdult = isAdult)
    }

    fun getNewContentByTheme(theme: String, member: Member, pageable: Pageable): List<GetAudioContentMainItem> {
        return repository.findByTheme(
            cloudfrontHost = imageHost,
            theme = theme,
            isAdult = member.auth != null,
            offset = pageable.offset,
            limit = pageable.pageSize.toLong()
        )
            .asSequence()
            .filter { !blockMemberRepository.isBlocked(blockedMemberId = member.id!!, memberId = it.creatorId) }
            .toList()
    }

    fun getNewContentFor2WeeksByTheme(theme: String, member: Member, pageable: Pageable): GetNewContentAllResponse {
        val totalCount = repository.totalCountNewContentFor2Weeks(theme, isAdult = member.auth != null)
        val items = repository.findByThemeFor2Weeks(
            cloudfrontHost = imageHost,
            theme = theme,
            isAdult = member.auth != null,
            offset = pageable.offset,
            limit = pageable.pageSize.toLong()
        )

        return GetNewContentAllResponse(totalCount, items)
    }

    @Transactional(readOnly = true)
    @Cacheable(cacheNames = ["default"], key = "'newContentUploadCreatorList:' + #memberId + ':' + #isAdult")
    fun getNewContentUploadCreatorList(memberId: Long, isAdult: Boolean): List<GetNewContentUploadCreator> {
        return repository.getNewContentUploadCreatorList(
            cloudfrontHost = imageHost,
            isAdult = isAdult
        )
            .asSequence()
            .filter { !blockMemberRepository.isBlocked(blockedMemberId = memberId, memberId = it.creatorId) }
            .toList()
    }

    @Transactional(readOnly = true)
    @Cacheable(cacheNames = ["default"], key = "'contentMainBannerList:' + #memberId + ':' + #isAdult")
    fun getAudioContentMainBannerList(memberId: Long, isAdult: Boolean) =
        repository.getAudioContentMainBannerList(isAdult = isAdult)
            .asSequence()
            .filter {
                if (it.type == AudioContentBannerType.CREATOR && it.creator != null) {
                    !blockMemberRepository.isBlocked(blockedMemberId = memberId, memberId = it.creator!!.id!!)
                } else {
                    true
                }
            }
            .map {
                GetAudioContentBannerResponse(
                    type = it.type,
                    thumbnailImageUrl = "$imageHost/${it.thumbnailImage}",
                    eventItem = if (it.type == AudioContentBannerType.EVENT && it.event != null) {
                        EventItem(
                            id = it.event!!.id!!,
                            thumbnailImageUrl = if (!it.event!!.thumbnailImage.startsWith("https://")) {
                                "$imageHost/${it.event!!.thumbnailImage}"
                            } else {
                                it.event!!.thumbnailImage
                            },
                            detailImageUrl = if (
                                it.event!!.detailImage != null &&
                                !it.event!!.detailImage!!.startsWith("https://")
                            ) {
                                "$imageHost/${it.event!!.detailImage}"
                            } else {
                                it.event!!.detailImage
                            },
                            popupImageUrl = null,
                            link = it.event!!.link,
                            title = it.event!!.title,
                            isPopup = false
                        )
                    } else {
                        null
                    },
                    creatorId = if (it.type == AudioContentBannerType.CREATOR && it.creator != null) {
                        it.creator!!.id
                    } else {
                        null
                    },
                    link = it.link
                )
            }
            .toList()

    @Transactional(readOnly = true)
    @Cacheable(cacheNames = ["default"], key = "'contentCurationList:' + #memberId + ':' + #isAdult")
    fun getAudioContentCurationList(memberId: Long, isAdult: Boolean) =
        repository.getAudioContentCurations(isAdult = isAdult)
            .asSequence()
            .map {
                GetAudioContentCurationResponse(
                    curationId = it.id!!,
                    title = it.title,
                    description = it.description,
                    contents = repository.findAudioContentByCurationId(
                        curationId = it.id!!,
                        cloudfrontHost = imageHost,
                        isAdult = isAdult
                    )
                        .asSequence()
                        .filter { content ->
                            !blockMemberRepository.isBlocked(blockedMemberId = memberId, memberId = content.creatorId)
                        }
                        .toList()
                )
            }
            .filter { it.contents.isNotEmpty() }
            .toList()
}
