package kr.co.vividnext.sodalive.creator.admin.calculate

import kr.co.vividnext.sodalive.admin.calculate.CumulativeSalesByContentItem
import kr.co.vividnext.sodalive.admin.calculate.GetCalculateContentDonationResponse
import kr.co.vividnext.sodalive.admin.calculate.GetCalculateContentResponse
import kr.co.vividnext.sodalive.admin.calculate.GetCalculateLiveResponse
import kr.co.vividnext.sodalive.admin.calculate.GetCumulativeSalesByContentResponse
import kr.co.vividnext.sodalive.can.use.CanUsage
import kr.co.vividnext.sodalive.content.order.OrderType
import kr.co.vividnext.sodalive.member.Member
import org.springframework.cache.annotation.Cacheable
import org.springframework.stereotype.Service
import org.springframework.transaction.annotation.Transactional
import java.time.LocalDate
import java.time.ZoneId
import java.time.format.DateTimeFormatter
import kotlin.math.roundToInt

@Service
class CreatorAdminCalculateService(private val repository: CreatorAdminCalculateQueryRepository) {
    @Transactional(readOnly = true)
    @Cacheable(
        cacheNames = ["default"],
        key = "'creatorCalculateLive:' + " + "#member + ':' + #startDateStr + ':' + #endDateStr"
    )
    fun getCalculateLive(startDateStr: String, endDateStr: String, member: Member): List<GetCalculateLiveResponse> {
        val dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd")
        val startDate = LocalDate.parse(startDateStr, dateTimeFormatter).atTime(0, 0, 0)
            .atZone(ZoneId.of("Asia/Seoul"))
            .withZoneSameInstant(ZoneId.of("UTC"))
            .toLocalDateTime()

        val endDate = LocalDate.parse(endDateStr, dateTimeFormatter).atTime(23, 59, 59)
            .atZone(ZoneId.of("Asia/Seoul"))
            .withZoneSameInstant(ZoneId.of("UTC"))
            .toLocalDateTime()

        return repository
            .getCalculateLive(startDate, endDate, member.id!!)
            .asSequence()
            .map {
                val canUsageStr = if (it.canUsage == CanUsage.LIVE) {
                    "유료"
                } else {
                    "후원"
                }

                val numberOfPeople = if (it.canUsage == CanUsage.LIVE) {
                    it.memberCount.toInt()
                } else {
                    0
                }

                // 원화 = totalCoin * 100 ( 캔 1개 = 100원 )
                val totalKrw = it.totalAmount * 100

                // 결제수수료 : 6.6%
                val paymentFee = totalKrw * 0.066f

                // 정산금액 = (원화 - 결제수수료) 의 70%
                val settlementAmount = (totalKrw.toFloat() - paymentFee) * 0.7

                // 원천세 = 정산금액의 3.3%
                val tax = settlementAmount * 0.033

                // 입금액
                val depositAmount = settlementAmount - tax

                GetCalculateLiveResponse(
                    email = it.email,
                    nickname = it.nickname,
                    date = it.date,
                    title = it.title,
                    entranceFee = it.entranceFee,
                    canUsageStr = canUsageStr,
                    numberOfPeople = numberOfPeople,
                    totalAmount = it.totalAmount,
                    totalKrw = totalKrw,
                    paymentFee = paymentFee.roundToInt(),
                    settlementAmount = settlementAmount.roundToInt(),
                    tax = tax.roundToInt(),
                    depositAmount = depositAmount.roundToInt()
                )
            }
            .toList()
    }

    @Transactional(readOnly = true)
    @Cacheable(
        cacheNames = ["cache_ttl_3_hours"],
        key = "'creatorCalculateContentList:' + " +
            "#startDateStr + ':' + #endDateStr + ':' + #memberId + ':' + #offset + ':' + #limit"
    )
    fun getCalculateContentList(
        startDateStr: String,
        endDateStr: String,
        memberId: Long,
        offset: Long,
        limit: Long
    ): GetCalculateContentListResponse {
        val dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd")
        val startDate = LocalDate.parse(startDateStr, dateTimeFormatter).atTime(0, 0, 0)
            .atZone(ZoneId.of("Asia/Seoul"))
            .withZoneSameInstant(ZoneId.of("UTC"))
            .toLocalDateTime()

        val endDate = LocalDate.parse(endDateStr, dateTimeFormatter).atTime(23, 59, 59)
            .atZone(ZoneId.of("Asia/Seoul"))
            .withZoneSameInstant(ZoneId.of("UTC"))
            .toLocalDateTime()

        val totalCount = repository.getCalculateContentListTotalCount(startDate, endDate, memberId)
        val items = repository.getCalculateContentList(startDate, endDate, memberId, offset, limit)
            .asSequence()
            .map {
                val orderTypeStr = if (it.orderType == OrderType.RENTAL) {
                    "대여"
                } else {
                    "소장"
                }

                // 원화 = totalCoin * 100 ( 캔 1개 = 100원 )
                val totalKrw = it.totalCan * 100

                // 결제수수료 : 6.6%
                val paymentFee = totalKrw * 0.066f

                // 정산금액 = (원화 - 결제수수료) 의 70%
                val settlementAmount = (totalKrw.toFloat() - paymentFee) * 0.7

                // 원천세 = 정산금액의 3.3%
                val tax = settlementAmount * 0.033

                // 입금액
                val depositAmount = settlementAmount - tax

                GetCalculateContentResponse(
                    nickname = it.nickname,
                    title = it.title,
                    registrationDate = it.registrationDate,
                    saleDate = it.saleDate,
                    orderType = orderTypeStr,
                    orderPrice = it.orderPrice,
                    numberOfPeople = it.numberOfPeople.toInt(),
                    totalCan = it.totalCan,
                    totalKrw = totalKrw,
                    paymentFee = paymentFee.roundToInt(),
                    settlementAmount = settlementAmount.roundToInt(),
                    tax = tax.roundToInt(),
                    depositAmount = depositAmount.roundToInt()
                )
            }
            .toList()

        return GetCalculateContentListResponse(totalCount, items)
    }

    @Transactional(readOnly = true)
    @Cacheable(
        cacheNames = ["cache_ttl_3_hours"],
        key = "'creatorCumulativeSalesByContent:' + " + "#memberId + ':' + #offset + ':' + #limit"
    )
    fun getCumulativeSalesByContent(memberId: Long, offset: Long, limit: Long): GetCumulativeSalesByContentResponse {
        val totalCount = repository.getCumulativeSalesByContentTotalCount(memberId)
        val items = repository
            .getCumulativeSalesByContent(memberId, offset, limit)
            .asSequence()
            .map {
                val orderTypeStr = if (it.orderType == OrderType.RENTAL) {
                    "대여"
                } else {
                    "소장"
                }

                // 원화 = totalCoin * 100 ( 캔 1개 = 100원 )
                val totalKrw = it.totalCan * 100

                // 결제수수료 : 6.6%
                val paymentFee = totalKrw * 0.066f

                // 정산금액 = (원화 - 결제수수료) 의 70%
                val settlementAmount = (totalKrw.toFloat() - paymentFee) * 0.7

                // 원천세 = 정산금액의 3.3%
                val tax = settlementAmount * 0.033

                // 입금액
                val depositAmount = settlementAmount - tax

                CumulativeSalesByContentItem(
                    nickname = it.nickname,
                    title = it.title,
                    registrationDate = it.registrationDate,
                    orderType = orderTypeStr,
                    orderPrice = it.orderPrice,
                    numberOfPeople = it.numberOfPeople.toInt(),
                    totalCan = it.totalCan,
                    totalKrw = totalKrw,
                    paymentFee = paymentFee.roundToInt(),
                    settlementAmount = settlementAmount.roundToInt(),
                    tax = tax.roundToInt(),
                    depositAmount = depositAmount.roundToInt()
                )
            }
            .toList()

        return GetCumulativeSalesByContentResponse(totalCount, items)
    }

    @Transactional(readOnly = true)
    @Cacheable(
        cacheNames = ["cache_ttl_3_hours"],
        key = "'creatorCalculateContentDonationList2:' + " +
            "#startDateStr + ':' + #endDateStr + ':' + #memberId + ':' + #offset + ':' + #limit"
    )
    fun getCalculateContentDonationList(
        startDateStr: String,
        endDateStr: String,
        memberId: Long,
        offset: Long,
        limit: Long
    ): GetCreatorCalculateContentDonationResponse {
        val dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd")
        val startDate = LocalDate.parse(startDateStr, dateTimeFormatter).atTime(0, 0, 0)
            .atZone(ZoneId.of("Asia/Seoul"))
            .withZoneSameInstant(ZoneId.of("UTC"))
            .toLocalDateTime()

        val endDate = LocalDate.parse(endDateStr, dateTimeFormatter).atTime(23, 59, 59)
            .atZone(ZoneId.of("Asia/Seoul"))
            .withZoneSameInstant(ZoneId.of("UTC"))
            .toLocalDateTime()

        val totalCount = repository.getCalculateContentDonationListTotalCount(startDate, endDate, memberId)
        val items = repository
            .getCalculateContentDonationList(startDate, endDate, memberId, offset, limit)
            .asSequence()
            .map {
                // 원화 = totalCoin * 100 ( 캔 1개 = 100원 )
                val totalKrw = it.totalCan * 100

                // 결제수수료 : 6.6%
                val paymentFee = totalKrw * 0.066f

                // 정산금액
                // 유료콘텐츠 (원화 - 결제수수료) 의 90%
                // 무료콘텐츠 (원화 - 결제수수료) 의 70%
                val settlementAmount = if (it.price > 0) {
                    (totalKrw.toFloat() - paymentFee) * 0.9
                } else {
                    (totalKrw.toFloat() - paymentFee) * 0.7
                }

                // 원천세 = 정산금액의 3.3%
                val tax = settlementAmount * 0.033

                // 입금액
                val depositAmount = settlementAmount - tax

                GetCalculateContentDonationResponse(
                    nickname = it.nickname,
                    title = it.title,
                    paidOrFree = if (it.price > 0) {
                        "유료"
                    } else {
                        "무료"
                    },
                    registrationDate = it.registrationDate,
                    donationDate = it.donationDate,
                    numberOfDonation = it.numberOfDonation.toInt(),
                    totalCan = it.totalCan,
                    totalKrw = totalKrw,
                    paymentFee = paymentFee.roundToInt(),
                    settlementAmount = settlementAmount.roundToInt(),
                    tax = tax.roundToInt(),
                    depositAmount = depositAmount.roundToInt()
                )
            }
            .toList()

        return GetCreatorCalculateContentDonationResponse(totalCount, items)
    }
}
