package kr.co.vividnext.sodalive.creator.admin.member

import kr.co.vividnext.sodalive.common.ApiResponse
import kr.co.vividnext.sodalive.common.SodaException
import kr.co.vividnext.sodalive.jwt.TokenProvider
import kr.co.vividnext.sodalive.member.MemberRepository
import kr.co.vividnext.sodalive.member.MemberRole
import kr.co.vividnext.sodalive.member.login.LoginRequest
import kr.co.vividnext.sodalive.member.login.LoginResponse
import kr.co.vividnext.sodalive.member.token.MemberTokenRepository
import org.springframework.beans.factory.annotation.Value
import org.springframework.data.repository.findByIdOrNull
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.stereotype.Service
import org.springframework.transaction.annotation.Transactional
import java.util.concurrent.locks.ReentrantReadWriteLock
import kotlin.concurrent.write

@Service
class CreatorAdminMemberService(
    private val repository: MemberRepository,
    private val tokenRepository: MemberTokenRepository,
    private val tokenProvider: TokenProvider,
    private val authenticationManagerBuilder: AuthenticationManagerBuilder,

    @Value("\${cloud.aws.cloud-front.host}")
    private val cloudFrontHost: String
) {

    private val tokenLocks: MutableMap<Long, ReentrantReadWriteLock> = mutableMapOf()

    fun login(request: LoginRequest): ApiResponse<LoginResponse> {
        return ApiResponse.ok(
            message = "로그인 되었습니다.",
            data = login(request.email, request.password)
        )
    }

    @Transactional
    fun logout(token: String, memberId: Long) {
        val member = repository.findByIdOrNull(memberId)
            ?: throw SodaException("로그인 정보를 확인해주세요.")

        member.pushToken = null

        val lock = getOrCreateLock(memberId = memberId)
        lock.write {
            val memberToken = tokenRepository.findByIdOrNull(memberId)
                ?: throw SodaException("로그인 정보를 확인해주세요.")

            memberToken.tokenSet.remove(token)
            tokenRepository.save(memberToken)
        }
    }

    private fun login(email: String, password: String): LoginResponse {
        val member = repository.findByEmail(email = email) ?: throw SodaException("로그인 정보를 확인해주세요.")
        if (!member.isActive) {
            throw SodaException("탈퇴한 계정입니다.\n고객센터로 문의해 주시기 바랍니다.")
        }

        if (member.role != MemberRole.CREATOR) {
            throw SodaException("로그인 정보를 확인해주세요.")
        }

        val authenticationToken = UsernamePasswordAuthenticationToken(email, password)
        val authentication = authenticationManagerBuilder.`object`.authenticate(authenticationToken)
        SecurityContextHolder.getContext().authentication = authentication

        val jwt = tokenProvider.createToken(
            authentication = authentication,
            memberId = member.id!!
        )

        return LoginResponse(
            userId = member.id!!,
            token = jwt,
            nickname = member.nickname,
            email = member.email,
            profileImage = if (member.profileImage != null) {
                "$cloudFrontHost/${member.profileImage}"
            } else {
                "$cloudFrontHost/profile/default-profile.png"
            }
        )
    }

    private fun getOrCreateLock(memberId: Long): ReentrantReadWriteLock {
        return tokenLocks.computeIfAbsent(memberId) { ReentrantReadWriteLock() }
    }
}
