package kr.co.vividnext.sodalive.live.room

import com.querydsl.core.types.OrderSpecifier
import com.querydsl.core.types.Predicate
import com.querydsl.core.types.Projections
import com.querydsl.core.types.dsl.CaseBuilder
import com.querydsl.core.types.dsl.Expressions
import com.querydsl.jpa.impl.JPAQueryFactory
import kr.co.vividnext.sodalive.can.use.CanUsage
import kr.co.vividnext.sodalive.can.use.QUseCan.useCan
import kr.co.vividnext.sodalive.can.use.QUseCanCalculate.useCanCalculate
import kr.co.vividnext.sodalive.live.room.QLiveRoom.liveRoom
import kr.co.vividnext.sodalive.live.room.donation.GetLiveRoomDonationItem
import kr.co.vividnext.sodalive.live.room.donation.QGetLiveRoomDonationItem
import kr.co.vividnext.sodalive.member.Member
import kr.co.vividnext.sodalive.member.QMember
import kr.co.vividnext.sodalive.member.QMember.member
import org.springframework.data.domain.Pageable
import org.springframework.data.jpa.repository.JpaRepository
import org.springframework.stereotype.Repository
import java.time.LocalDate
import java.time.LocalDateTime
import java.time.ZoneId
import java.time.format.DateTimeFormatter

@Repository
interface LiveRoomRepository : JpaRepository<LiveRoom, Long>, LiveRoomQueryRepository

interface LiveRoomQueryRepository {
    fun getLiveRoomList(
        dateString: String?,
        status: LiveRoomStatus,
        pageable: Pageable,
        member: Member,
        timezone: String,
        isAdult: Boolean
    ): List<LiveRoom>

    fun getLiveRoom(id: Long): LiveRoom?
    fun getLiveRoomAndAccountId(roomId: Long, memberId: Long): LiveRoom?
    fun getRecentRoomInfo(memberId: Long, cloudFrontHost: String): GetRecentRoomInfoResponse?
    fun getDonationTotal(roomId: Long): Int?
    fun getDonationList(roomId: Long, cloudFrontHost: String): List<GetLiveRoomDonationItem>
}

class LiveRoomQueryRepositoryImpl(private val queryFactory: JPAQueryFactory) : LiveRoomQueryRepository {
    override fun getLiveRoomList(
        dateString: String?,
        status: LiveRoomStatus,
        pageable: Pageable,
        member: Member,
        timezone: String,
        isAdult: Boolean
    ): List<LiveRoom> {
        var where: Predicate

        if (status == LiveRoomStatus.NOW) {
            where = liveRoom.channelName.isNotNull
                .and(liveRoom.channelName.isNotEmpty)
        } else {
            where = if (dateString != null) {
                val dateTimeFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd")
                val date = LocalDate.parse(dateString, dateTimeFormatter).atStartOfDay()
                    .atZone(ZoneId.of(timezone))
                    .withZoneSameInstant(ZoneId.of("UTC"))
                    .toLocalDateTime()

                liveRoom.beginDateTime.goe(date)
                    .and(liveRoom.beginDateTime.lt(date.plusDays(1)))
                    .and(
                        liveRoom.channelName.isNull
                            .or(liveRoom.channelName.isEmpty)
                    )
            } else {
                liveRoom.beginDateTime.gt(
                    LocalDateTime.now()
                        .atZone(ZoneId.of(timezone))
                        .withZoneSameInstant(ZoneId.of("UTC"))
                        .toLocalDateTime()
                )
                    .and(
                        liveRoom.channelName.isNull
                            .or(liveRoom.channelName.isEmpty)
                    )
            }
        }

        if (!isAdult) {
            where = where.and(liveRoom.isAdult.isFalse)
        }

        where = where.and(liveRoom.isActive.isTrue)
            .and(liveRoom.member.isNotNull)

        return queryFactory
            .selectFrom(liveRoom)
            .innerJoin(liveRoom.member, QMember.member)
            .offset(pageable.offset)
            .limit(pageable.pageSize.toLong())
            .where(where)
            .orderBy(
                *orderByFieldAccountId(
                    memberId = member.id!!,
                    status = status,
                    offset = pageable.offset,
                    dateString = dateString
                )
            )
            .fetch()
    }

    override fun getLiveRoom(id: Long): LiveRoom? {
        return queryFactory
            .selectFrom(liveRoom)
            .innerJoin(liveRoom.member, member)
            .where(
                liveRoom.id.eq(id)
                    .and(liveRoom.isActive.isTrue)
            )
            .fetchFirst()
    }

    override fun getLiveRoomAndAccountId(roomId: Long, memberId: Long): LiveRoom? {
        return queryFactory
            .selectFrom(liveRoom)
            .innerJoin(liveRoom.member, member)
            .where(
                liveRoom.id.eq(roomId)
                    .and(liveRoom.isActive.isTrue)
                    .and(member.id.eq(memberId))
            )
            .fetchFirst()
    }

    override fun getRecentRoomInfo(memberId: Long, cloudFrontHost: String): GetRecentRoomInfoResponse? {
        return queryFactory
            .select(
                Projections.constructor(
                    GetRecentRoomInfoResponse::class.java,
                    liveRoom.title,
                    liveRoom.notice,
                    liveRoom.coverImage.prepend("/").prepend(cloudFrontHost),
                    liveRoom.coverImage,
                    liveRoom.numberOfPeople
                )
            )
            .from(liveRoom)
            .where(liveRoom.member.id.eq(memberId))
            .orderBy(liveRoom.id.desc())
            .limit(1)
            .fetchFirst()
    }

    override fun getDonationTotal(roomId: Long): Int? {
        return queryFactory
            .select(useCanCalculate.can.sum())
            .from(useCanCalculate)
            .innerJoin(useCanCalculate.useCan, useCan)
            .innerJoin(useCan.room, liveRoom)
            .where(
                liveRoom.id.eq(roomId)
                    .and(useCan.canUsage.eq(CanUsage.DONATION))
                    .and(useCan.isRefund.isFalse)
            )
            .fetchOne()
    }

    override fun getDonationList(roomId: Long, cloudFrontHost: String): List<GetLiveRoomDonationItem> {
        return queryFactory
            .select(
                QGetLiveRoomDonationItem(
                    member.profileImage
                        .coalesce("profile/default-profile.png")
                        .prepend("/")
                        .prepend(cloudFrontHost),
                    member.nickname,
                    member.id.coalesce(0),
                    useCan.can.sum().add(useCan.rewardCan.sum())
                )
            )
            .from(useCan)
            .join(useCan.member, member)
            .groupBy(useCan.member)
            .where(
                useCan.room.id.eq(roomId)
                    .and(useCan.canUsage.eq(CanUsage.DONATION))
                    .and(useCan.isRefund.isFalse)
            )
            .orderBy(useCan.can.sum().add(useCan.rewardCan.sum()).desc())
            .fetch()
    }

    private fun orderByFieldAccountId(
        memberId: Long,
        status: LiveRoomStatus,
        offset: Long,
        dateString: String?
    ): Array<out OrderSpecifier<*>> {
        return if (status == LiveRoomStatus.NOW) {
            arrayOf(Expressions.numberTemplate(Double::class.java, "function('rand')").asc())
        } else if (status == LiveRoomStatus.RESERVATION && offset == 0L && dateString == null) {
            arrayOf(
                CaseBuilder()
                    .`when`(member.id.eq(memberId)).then(1)
                    .otherwise(2)
                    .asc(),
                liveRoom.beginDateTime.asc()
            )
        } else {
            arrayOf(liveRoom.beginDateTime.asc())
        }
    }
}
