package kr.co.vividnext.sodalive.live.room

import com.amazonaws.services.s3.model.ObjectMetadata
import com.fasterxml.jackson.databind.ObjectMapper
import kr.co.vividnext.sodalive.agora.RtcTokenBuilder
import kr.co.vividnext.sodalive.agora.RtmTokenBuilder
import kr.co.vividnext.sodalive.aws.s3.S3Uploader
import kr.co.vividnext.sodalive.can.CanRepository
import kr.co.vividnext.sodalive.can.charge.Charge
import kr.co.vividnext.sodalive.can.charge.ChargeRepository
import kr.co.vividnext.sodalive.can.charge.ChargeStatus
import kr.co.vividnext.sodalive.can.payment.CanPaymentService
import kr.co.vividnext.sodalive.can.payment.Payment
import kr.co.vividnext.sodalive.can.payment.PaymentGateway
import kr.co.vividnext.sodalive.can.payment.PaymentStatus
import kr.co.vividnext.sodalive.can.use.CanUsage
import kr.co.vividnext.sodalive.can.use.UseCanCalculateRepository
import kr.co.vividnext.sodalive.can.use.UseCanCalculateStatus
import kr.co.vividnext.sodalive.common.SodaException
import kr.co.vividnext.sodalive.explorer.ExplorerQueryRepository
import kr.co.vividnext.sodalive.extensions.convertLocalDateTime
import kr.co.vividnext.sodalive.fcm.FcmEvent
import kr.co.vividnext.sodalive.fcm.FcmEventType
import kr.co.vividnext.sodalive.live.reservation.LiveReservationRepository
import kr.co.vividnext.sodalive.live.room.cancel.CancelLiveRequest
import kr.co.vividnext.sodalive.live.room.cancel.LiveRoomCancel
import kr.co.vividnext.sodalive.live.room.cancel.LiveRoomCancelRepository
import kr.co.vividnext.sodalive.live.room.detail.GetRoomDetailManager
import kr.co.vividnext.sodalive.live.room.detail.GetRoomDetailResponse
import kr.co.vividnext.sodalive.live.room.detail.GetRoomDetailUser
import kr.co.vividnext.sodalive.live.room.donation.DeleteLiveRoomDonationMessage
import kr.co.vividnext.sodalive.live.room.donation.GetLiveRoomDonationStatusResponse
import kr.co.vividnext.sodalive.live.room.donation.GetLiveRoomDonationTotalResponse
import kr.co.vividnext.sodalive.live.room.donation.LiveRoomDonationMessage
import kr.co.vividnext.sodalive.live.room.donation.LiveRoomDonationRequest
import kr.co.vividnext.sodalive.live.room.info.GetRoomInfoResponse
import kr.co.vividnext.sodalive.live.room.info.LiveRoomInfo
import kr.co.vividnext.sodalive.live.room.info.LiveRoomInfoRedisRepository
import kr.co.vividnext.sodalive.live.room.info.LiveRoomMember
import kr.co.vividnext.sodalive.live.room.kickout.LiveRoomKickOutService
import kr.co.vividnext.sodalive.live.room.visit.LiveRoomVisitService
import kr.co.vividnext.sodalive.live.tag.LiveTagRepository
import kr.co.vividnext.sodalive.member.Gender
import kr.co.vividnext.sodalive.member.Member
import kr.co.vividnext.sodalive.member.MemberRepository
import kr.co.vividnext.sodalive.member.MemberRole
import kr.co.vividnext.sodalive.member.block.BlockMemberRepository
import kr.co.vividnext.sodalive.utils.generateFileName
import org.springframework.beans.factory.annotation.Value
import org.springframework.context.ApplicationEventPublisher
import org.springframework.data.domain.Pageable
import org.springframework.data.repository.findByIdOrNull
import org.springframework.stereotype.Service
import org.springframework.transaction.annotation.Transactional
import org.springframework.web.multipart.MultipartFile
import java.time.LocalDateTime
import java.time.ZoneId
import java.time.format.DateTimeFormatter
import java.util.Date
import java.util.Locale
import java.util.concurrent.locks.ReentrantReadWriteLock
import kotlin.concurrent.write

@Service
@Transactional(readOnly = true)
class LiveRoomService(
    private val repository: LiveRoomRepository,
    private val roomInfoRepository: LiveRoomInfoRedisRepository,
    private val roomCancelRepository: LiveRoomCancelRepository,
    private val kickOutService: LiveRoomKickOutService,
    private val blockMemberRepository: BlockMemberRepository,

    private val applicationEventPublisher: ApplicationEventPublisher,
    private val useCanCalculateRepository: UseCanCalculateRepository,
    private val reservationRepository: LiveReservationRepository,
    private val explorerQueryRepository: ExplorerQueryRepository,
    private val roomVisitService: LiveRoomVisitService,
    private val canPaymentService: CanPaymentService,
    private val chargeRepository: ChargeRepository,
    private val memberRepository: MemberRepository,
    private val tagRepository: LiveTagRepository,
    private val canRepository: CanRepository,
    private val objectMapper: ObjectMapper,
    private val s3Uploader: S3Uploader,

    private val rtcTokenBuilder: RtcTokenBuilder,
    private val rtmTokenBuilder: RtmTokenBuilder,

    @Value("\${agora.app-id}")
    private val agoraAppId: String,

    @Value("\${agora.app-certificate}")
    private val agoraAppCertificate: String,

    @Value("\${cloud.aws.s3.bucket}")
    private val coverImageBucket: String,
    @Value("\${cloud.aws.cloud-front.host}")
    private val cloudFrontHost: String
) {
    private val tokenLocks: MutableMap<Long, ReentrantReadWriteLock> = mutableMapOf()

    fun getRoomList(
        dateString: String?,
        status: LiveRoomStatus,
        pageable: Pageable,
        member: Member,
        timezone: String
    ): List<GetRoomListResponse> {
        return repository
            .getLiveRoomList(
                dateString = dateString,
                status = status,
                pageable = pageable,
                member = member,
                timezone = timezone,
                isAdult = member.auth != null
            )
            .asSequence()
            .filter { !blockMemberRepository.isBlocked(blockedMemberId = member.id!!, memberId = it.member!!.id!!) }
            .map {
                val roomInfo = roomInfoRepository.findByIdOrNull(it.id!!)

                val reservations = it.reservations
                    .filter { reservation -> reservation.member!!.id!! == member.id!! && reservation.isActive }

                val beginDateTime = it.beginDateTime
                    .atZone(ZoneId.of("UTC"))
                    .withZoneSameInstant(ZoneId.of(timezone))
                    .format(
                        DateTimeFormatter
                            .ofPattern("yyyy년 MM월 dd일 (E) a hh시 mm분")
                            .withLocale(Locale.KOREAN)
                    )

                GetRoomListResponse(
                    roomId = it.id!!,
                    title = it.title,
                    content = it.notice,
                    beginDateTime = beginDateTime,
                    numberOfParticipate = (roomInfo?.listenerCount ?: 0) +
                        (roomInfo?.speakerCount ?: 0) +
                        (roomInfo?.managerCount ?: 0),
                    numberOfPeople = it.numberOfPeople,
                    isAdult = it.isAdult,
                    price = it.price,
                    channelName = it.channelName,
                    creatorNickname = it.member!!.nickname,
                    creatorId = it.member!!.id!!,
                    tags = it.tags
                        .asSequence()
                        .filter { tag -> tag.tag.isActive }
                        .map { tag -> tag.tag.tag }
                        .toList(),
                    coverImageUrl = if (it.coverImage!!.startsWith("https://")) {
                        it.coverImage!!
                    } else {
                        "$cloudFrontHost/${it.coverImage!!}"
                    },
                    isReservation = reservations.isNotEmpty(),
                    isPrivateRoom = it.type == LiveRoomType.PRIVATE
                )
            }
            .toList()
    }

    @Transactional
    fun createLiveRoom(coverImage: MultipartFile?, requestString: String, member: Member): CreateLiveRoomResponse {
        val request = objectMapper.readValue(requestString, CreateLiveRoomRequest::class.java)
        if (request.coverImageUrl == null && coverImage == null) {
            throw SodaException("커버이미지를 선택해 주세요.")
        }

        val now = LocalDateTime.now()
        val beginDateTime = if (request.beginDateTimeString != null) {
            request.beginDateTimeString.convertLocalDateTime("yyyy-MM-dd HH:mm")
                .atZone(ZoneId.of(request.timezone))
                .withZoneSameInstant(ZoneId.of("UTC"))
                .toLocalDateTime()
        } else {
            now
        }

        if (
            request.beginDateTimeString != null &&
            beginDateTime < now.plusMinutes(30)
        ) {
            throw SodaException("현재시각 기준, 30분 이후부터 설정가능합니다.")
        }

        if (
            request.type == LiveRoomType.PRIVATE &&
            (request.password == null || request.password.length != 6)
        ) {
            throw SodaException("방 입장 비밀번호 6자리를 입력해 주세요.")
        }

        if (request.price in 1..9) {
            throw SodaException("유료라이브는 10캔부터 설정 가능 합니다.")
        }

        val room = LiveRoom(
            title = request.title,
            notice = request.content,
            beginDateTime = beginDateTime,
            numberOfPeople = request.numberOfPeople,
            isAdult = request.isAdult,
            price = if (request.price < 0) {
                0
            } else {
                request.price
            },
            type = request.type,
            password = request.password
        )
        room.member = member

        if (request.beginDateTimeString == null) {
            room.channelName = "SODA_LIVE_CHANNEL_" +
                "${member.id!!}_${beginDateTime.year}_${beginDateTime.month}_${beginDateTime.dayOfMonth}_" +
                "${beginDateTime.hour}_${beginDateTime.minute}"
        }

        request.tags.forEach {
            val tag = tagRepository.findByTag(it)
            if (tag != null) {
                room.tags.add(LiveRoomTag(room, tag))

                if (tag.tag.contains("음담패설")) {
                    room.isAdult = true
                }
            }
        }

        val createdRoom = repository.save(room)
        // 이미지 업로드
        if (coverImage != null) {
            val metadata = ObjectMetadata()
            metadata.contentLength = coverImage.size

            // 커버 이미지 파일명 생성
            val coverImageFileName = generateFileName(prefix = "${room.id}-cover")

            // 커버 이미지 업로드
            val coverImagePath = s3Uploader.upload(
                inputStream = coverImage.inputStream,
                bucket = coverImageBucket,
                filePath = "live_room_cover/${room.id}/$coverImageFileName",
                metadata = metadata
            )

            room.coverImage = coverImagePath
            room.bgImage = coverImagePath
        } else {
            room.coverImage = request.coverImageUrl
            room.bgImage = request.coverImageUrl
        }

        applicationEventPublisher.publishEvent(
            FcmEvent(
                type = FcmEventType.CREATE_LIVE,
                title = createdRoom.member!!.nickname,
                message = if (createdRoom.channelName != null) {
                    "라이브를 시작했습니다. - ${createdRoom.title}"
                } else {
                    "라이브를 예약했습니다. - ${createdRoom.title}"
                },
                isAuth = createdRoom.isAdult,
                roomId = createdRoom.id,
                creatorId = createdRoom.member!!.id,
                container = "ios"
            )
        )

        applicationEventPublisher.publishEvent(
            FcmEvent(
                type = FcmEventType.CREATE_LIVE,
                title = createdRoom.member!!.nickname,
                message = if (createdRoom.channelName != null) {
                    "라이브를 시작했습니다. - ${createdRoom.title}"
                } else {
                    "라이브를 예약했습니다. - ${createdRoom.title}"
                },
                isAuth = createdRoom.isAdult,
                roomId = createdRoom.id,
                creatorId = createdRoom.member!!.id,
                container = "aos"
            )
        )

        return CreateLiveRoomResponse(createdRoom.id, createdRoom.channelName)
    }

    fun getRoomDetail(roomId: Long, member: Member, timezone: String): GetRoomDetailResponse {
        val room = repository.getLiveRoom(id = roomId)
            ?: throw SodaException("이미 종료된 방입니다")

        if (room.isAdult && member.auth == null) {
            throw SodaException("본인인증이 필요한 서비스 입니다.\n소다라이브 마이페이지에서 본인인증 후 다시 이용해 주세요.")
        }

        val beginDateTime = room.beginDateTime
            .atZone(ZoneId.of("UTC"))
            .withZoneSameInstant(ZoneId.of(timezone))
            .format(DateTimeFormatter.ofPattern("yyyy.MM.dd E hh:mm a"))

        val response = GetRoomDetailResponse(
            roomId = roomId,
            title = room.title,
            notice = room.notice,
            price = room.price,
            tags = room.tags.asSequence().filter { it.tag.isActive }.map { it.tag.tag }.toList(),
            numberOfParticipantsTotal = room.numberOfPeople,
            numberOfParticipants = 0,
            channelName = room.channelName,
            beginDateTime = beginDateTime,
            isPaid = false,
            isAdult = room.isAdult,
            isPrivateRoom = room.type == LiveRoomType.PRIVATE,
            password = room.password
        )
        response.manager = GetRoomDetailManager(room.member!!, cloudFrontHost = cloudFrontHost)

        if (!room.channelName.isNullOrBlank()) {
            val roomInfo = roomInfoRepository.findByIdOrNull(roomId)

            if (roomInfo != null) {
                response.isPaid = canRepository.isExistPaidLiveRoom(
                    memberId = member.id!!,
                    roomId = roomId
                ) != null

                val users = roomInfo.speakerList
                    .asSequence()
                    .map {
                        GetRoomDetailUser(
                            it.id,
                            it.nickname,
                            "$cloudFrontHost/${it.profileImage}"
                        )
                    }.toMutableList()

                users.addAll(
                    roomInfo.listenerList
                        .asSequence()
                        .map {
                            GetRoomDetailUser(
                                it.id,
                                it.nickname,
                                "$cloudFrontHost/${it.profileImage}"
                            )
                        }
                )

                users.addAll(
                    roomInfo.managerList
                        .asSequence()
                        .map {
                            GetRoomDetailUser(
                                it.id,
                                it.nickname,
                                "$cloudFrontHost/${it.profileImage}"
                            )
                        }
                )

                response.participatingUsers = users
                response.numberOfParticipants = users.size
            }
        } else {
            val reservationList = reservationRepository.getReservationList(roomId)
            response.participatingUsers = reservationList
                .asSequence()
                .map {
                    if (it.member!!.id!! == member.id!!) {
                        response.isPaid = true
                    }

                    GetRoomDetailUser(it.member!!, cloudFrontHost)
                }
                .toList()
            response.numberOfParticipants = reservationList.size
        }

        return response
    }

    @Transactional
    fun startLive(request: StartLiveRequest, member: Member) {
        val room = repository.getLiveRoomAndAccountId(request.roomId, member.id!!)
            ?: throw SodaException("해당하는 라이브가 없습니다.")

        val nowDateTime = LocalDateTime.now()

        if (nowDateTime.plusMinutes(10).isBefore(room.beginDateTime)) {
            val startAvailableDateTimeString = room.beginDateTime
                .minusMinutes(10)
                .atZone(ZoneId.of("UTC"))
                .withZoneSameInstant(ZoneId.of("Asia/Seoul"))
                .format(DateTimeFormatter.ofPattern("yyyy년 MM월 dd일 (E) a hh시 mm분").withLocale(Locale.KOREAN))

            throw SodaException("$startAvailableDateTimeString 이후에 시작할 수 있습니다.")
        }

        val dateTime = nowDateTime
            .atZone(ZoneId.of("UTC"))
            .withZoneSameInstant(ZoneId.of(request.timezone))
            .toLocalDateTime()

        room.channelName = "SODA_LIVE_CHANNEL_" +
            "${member.id}_${dateTime.year}_${dateTime.month}_${dateTime.dayOfMonth}_" +
            "${dateTime.hour}_${dateTime.minute}"

        room.beginDateTime = nowDateTime

        applicationEventPublisher.publishEvent(
            FcmEvent(
                type = FcmEventType.START_LIVE,
                title = room.member!!.nickname,
                message = "라이브를 시작했습니다 - ${room.title}",
                isAuth = room.isAdult,
                roomId = room.id,
                creatorId = room.member!!.id,
                container = "ios"
            )
        )

        applicationEventPublisher.publishEvent(
            FcmEvent(
                type = FcmEventType.START_LIVE,
                title = room.member!!.nickname,
                message = "라이브를 시작했습니다 - ${room.title}",
                isAuth = room.isAdult,
                roomId = room.id,
                creatorId = room.member!!.id,
                container = "aos"
            )
        )
    }

    @Transactional
    fun cancelLive(request: CancelLiveRequest, member: Member) {
        val room = repository.getLiveRoomAndAccountId(request.roomId, member.id!!)
            ?: throw SodaException("해당하는 라이브가 없습니다.")

        if (request.reason.isBlank()) {
            throw SodaException("취소사유를 입력해 주세요.")
        }
        room.isActive = false

        val roomCancel = LiveRoomCancel(request.reason)
        roomCancel.room = room
        roomCancelRepository.save(roomCancel)

        // 유료방인 경우 환불처리
        if (room.price > 0) {
            val bookerList = reservationRepository.getReservationBookerList(roomId = room.id!!)
            for (booker in bookerList) {
                val useCan = canRepository.getCanUsedForLiveRoomNotRefund(
                    memberId = booker.id!!,
                    roomId = room.id!!,
                    canUsage = CanUsage.LIVE
                ) ?: continue
                useCan.isRefund = true

                val useCanCalculate = useCanCalculateRepository.findByUseCanIdAndStatus(useCanId = useCan.id!!)
                useCanCalculate.forEach {
                    it.status = UseCanCalculateStatus.REFUND

                    val charge = Charge(0, it.can, status = ChargeStatus.REFUND_CHARGE)
                    charge.title = "${it.can} 캔"
                    charge.useCan = useCan

                    when (it.paymentGateway) {
                        PaymentGateway.GOOGLE_IAP -> booker.googleRewardCan += charge.rewardCan
                        PaymentGateway.APPLE_IAP -> booker.appleRewardCan += charge.rewardCan
                        else -> booker.pgRewardCan += charge.rewardCan
                    }
                    charge.member = booker

                    val payment = Payment(
                        status = PaymentStatus.COMPLETE,
                        paymentGateway = it.paymentGateway
                    )
                    payment.method = "환불"
                    charge.payment = payment

                    chargeRepository.save(charge)
                }
            }
        }

        val pushTokenListMap = memberRepository.getPushTokenFromReservationList(request.roomId)
        reservationRepository.cancelReservation(roomId = room.id!!)

        applicationEventPublisher.publishEvent(
            FcmEvent(
                type = FcmEventType.CANCEL_LIVE,
                title = room.member!!.nickname,
                message = "라이브 취소 : ${room.title}",
                recipientsMap = pushTokenListMap
            )
        )
    }

    @Transactional
    fun enterLive(request: EnterOrQuitLiveRoomRequest, member: Member) {
        val room = repository.getLiveRoom(id = request.roomId)
            ?: throw SodaException("해당하는 라이브가 없습니다.")

        if (
            room.member!!.id!! != member.id!! &&
            room.type == LiveRoomType.PRIVATE &&
            (request.password == null || request.password != room.password)
        ) {
            throw SodaException("비밀번호가 일치하지 않습니다.\n다시 확인 후 입력해주세요.")
        }

        val isBlocked = blockMemberRepository.isBlocked(blockedMemberId = member.id!!, memberId = room.member!!.id!!)
        if (isBlocked) throw SodaException("${room.member!!.nickname}님의 요청으로 라이브에 입장할 수 없습니다.")

        val kickOutCount = kickOutService.getKickOutCount(roomId = room.id!!, userId = member.id!!)
        if (kickOutCount >= 2) throw SodaException("${room.member!!.nickname}님의 요청으로 라이브에 참여할 수 없습니다.")

        val lock = getOrCreateLock(memberId = member.id!!)
        lock.write {
            var roomInfo = roomInfoRepository.findByIdOrNull(request.roomId)
            if (roomInfo == null) {
                roomInfo = roomInfoRepository.save(LiveRoomInfo(roomId = request.roomId))
            }

            if (roomInfo.speakerCount + roomInfo.listenerCount + roomInfo.managerCount >= room.numberOfPeople) {
                throw SodaException("방이 가득찼습니다.")
            }

            if (
                room.price > 0 &&
                room.member!!.id!! != member.id!! &&
                canRepository.isExistPaidLiveRoom(memberId = member.id!!, roomId = request.roomId) == null
            ) {
                val findMember = memberRepository.findByIdOrNull(id = member.id!!)
                    ?: throw SodaException("로그인 정보를 확인해 주세요.")

                val totalCan = findMember.getChargeCan(request.container) + findMember.getRewardCan(request.container)
                if (totalCan < room.price) {
                    throw SodaException("${room.price - totalCan}캔이 부족합니다. 충전 후 이용해 주세요.")
                }

                canPaymentService.spendCan(
                    memberId = member.id!!,
                    needCan = room.price,
                    canUsage = CanUsage.LIVE,
                    liveRoom = room,
                    container = request.container
                )
            }

            roomInfo.removeListener(member)
            roomInfo.removeSpeaker(member)
            roomInfo.removeManager(member)

            if (room.member!!.id == member.id) {
                roomInfo.addSpeaker(member, cloudFrontHost)
            } else {
                roomInfo.addListener(member, cloudFrontHost)
            }

            roomInfoRepository.save(roomInfo)
            roomVisitService.roomVisit(room, member)
        }
    }

    fun getRecentRoomInfo(member: Member): GetRecentRoomInfoResponse {
        return repository.getRecentRoomInfo(memberId = member.id!!, cloudFrontHost = cloudFrontHost)
            ?: throw SodaException("최근 데이터가 없습니다.")
    }

    @Transactional
    fun editLiveRoomInfo(roomId: Long, coverImage: MultipartFile?, requestString: String?, member: Member) {
        val room = repository.getLiveRoom(roomId)
        if (member.id == null || room?.member?.id != member.id!!) {
            throw SodaException("잘못된 요청입니다.")
        }

        if (coverImage == null && requestString == null) {
            throw SodaException("변경사항이 없습니다.")
        }

        if (requestString != null) {
            val request = objectMapper.readValue(requestString, EditLiveRoomInfoRequest::class.java)

            if (request.title != null) {
                room.title = request.title
            }

            if (request.notice != null) {
                room.notice = request.notice
            }

            if (request.numberOfPeople != null) {
                room.numberOfPeople = request.numberOfPeople
            }

            if (request.beginDateTimeString != null && request.timezone != null) {
                room.beginDateTime = request.beginDateTimeString.convertLocalDateTime("yyyy-MM-dd HH:mm")
                    .atZone(ZoneId.of(request.timezone))
                    .withZoneSameInstant(ZoneId.of("UTC"))
                    .toLocalDateTime()
            }
        }

        if (coverImage != null) {
            val metadata = ObjectMetadata()
            metadata.contentLength = coverImage.size

            // 커버 이미지 파일명 생성
            val coverImageFileName = generateFileName(prefix = "${room.id}-cover")

            // 커버 이미지 업로드
            val coverImagePath = s3Uploader.upload(
                inputStream = coverImage.inputStream,
                bucket = coverImageBucket,
                filePath = "live_room_cover/${room.id}/$coverImageFileName",
                metadata = metadata
            )

            room.bgImage = coverImagePath

            if (room.channelName == null) {
                room.coverImage = coverImagePath
            }
        }
    }

    fun getRoomInfo(roomId: Long, member: Member): GetRoomInfoResponse {
        val roomInfo = roomInfoRepository.findByIdOrNull(roomId)
            ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

        val room = repository.findByIdOrNull(roomId)
            ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

        val currentTimeStamp = Date().time
        val expireTimestamp = (currentTimeStamp + (60 * 60 * 24 * 1000)) / 1000

        val rtcToken = rtcTokenBuilder.buildTokenWithUid(
            agoraAppId,
            agoraAppCertificate,
            room.channelName!!,
            member.id!!.toInt(),
            expireTimestamp.toInt()
        )

        val rtmToken = rtmTokenBuilder.buildToken(
            agoraAppId,
            agoraAppCertificate,
            member.id!!.toString(),
            expireTimestamp.toInt()
        )

        val isFollowing = explorerQueryRepository
            .getNotificationUserIds(room.member!!.id!!)
            .contains(member.id)

        val donationRankingTop3UserIds = if (room.member!!.isVisibleDonationRank) {
            explorerQueryRepository
                .getMemberDonationRanking(
                    room.member!!.id!!,
                    3,
                    withDonationCan = false
                )
                .asSequence()
                .map { it.userId }
                .toList()
        } else {
            listOf()
        }

        return GetRoomInfoResponse(
            roomId = roomId,
            title = room.title,
            notice = room.notice,
            coverImageUrl = if (room.bgImage != null) {
                if (room.bgImage!!.startsWith("https://")) {
                    room.bgImage!!
                } else {
                    "$cloudFrontHost/${room.bgImage!!}"
                }
            } else {
                if (room.coverImage!!.startsWith("https://")) {
                    room.coverImage!!
                } else {
                    "$cloudFrontHost/${room.coverImage!!}"
                }
            },
            channelName = room.channelName!!,
            rtcToken = rtcToken,
            rtmToken = rtmToken,
            creatorId = room.member!!.id!!,
            creatorNickname = room.member!!.nickname,
            creatorProfileUrl = if (room.member!!.profileImage != null) {
                "$cloudFrontHost/${room.member!!.profileImage}"
            } else {
                "$cloudFrontHost/profile/default-profile.png"
            },
            isFollowing = isFollowing,
            isAdult = room.isAdult,
            participantsCount = roomInfo.listenerCount + roomInfo.speakerCount + roomInfo.managerCount,
            totalAvailableParticipantsCount = room.numberOfPeople,
            speakerList = roomInfo.speakerList,
            listenerList = roomInfo.listenerList,
            managerList = roomInfo.managerList,
            donationRankingTop3UserIds = donationRankingTop3UserIds,
            isPrivateRoom = room.type == LiveRoomType.PRIVATE,
            password = room.password
        )
    }

    fun getDonationMessageList(roomId: Long, member: Member): List<LiveRoomDonationMessage> {
        val room = repository.findByIdOrNull(roomId)
            ?: throw SodaException("해당하는 라이브가 없습니다.")

        if (member.id!! != room.member!!.id!!) {
            throw SodaException("잘못된 요청입니다.")
        }

        val roomInfo = roomInfoRepository.findByIdOrNull(roomId)
            ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

        return roomInfo.donationMessageList
    }

    fun deleteDonationMessage(request: DeleteLiveRoomDonationMessage, member: Member) {
        val lock = getOrCreateLock(memberId = member.id!!)
        lock.write {
            val room = repository.findByIdOrNull(request.roomId)
                ?: throw SodaException("해당하는 라이브가 없습니다.")

            if (member.id!! != room.member!!.id!!) {
                throw SodaException("잘못된 요청입니다.")
            }

            val roomInfo = roomInfoRepository.findByIdOrNull(request.roomId)
                ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

            roomInfo.removeDonationMessage(request.messageUUID)
            roomInfoRepository.save(roomInfo)
        }
    }

    fun getUserProfile(roomId: Long, userId: Long, member: Member): GetLiveRoomUserProfileResponse {
        val room = repository.getLiveRoom(roomId)
            ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")
        val roomInfo = roomInfoRepository.findByIdOrNull(roomId)
            ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

        val user = memberRepository.findByIdOrNull(userId)
            ?: throw SodaException("잘못된 요청입니다.")

        val isFollowing = if (user.role == MemberRole.CREATOR) {
            explorerQueryRepository
                .getNotificationUserIds(userId)
                .contains(member.id!!)
        } else {
            null
        }

        // 조회 하는 유저
        val memberResponse = LiveRoomMember(member, cloudFrontHost)
        // 조회 당하는 유저
        val userResponse = LiveRoomMember(user, cloudFrontHost)

        val isSpeaker = if (
            room.member!!.id!! != userId &&
            (room.member!!.id!! == member.id!! || roomInfo.managerList.contains(memberResponse))
        ) {
            roomInfo.speakerList.contains(userResponse)
        } else {
            null
        }

        val isManager = if (room.member!!.id!! != userId && room.member!!.id!! == member.id!!) {
            roomInfo.managerList.contains(userResponse)
        } else {
            null
        }

        return GetLiveRoomUserProfileResponse(
            userId = user.id!!,
            nickname = user.nickname,
            profileUrl = if (user.profileImage != null) {
                "$cloudFrontHost/${user.profileImage}"
            } else {
                "$cloudFrontHost/profile/default-profile.png"
            },
            gender = if (user.gender == Gender.FEMALE) "여" else if (user.gender == Gender.MALE) "남" else "미",
            instagramUrl = user.instagramUrl,
            youtubeUrl = user.youtubeUrl,
            websiteUrl = user.websiteUrl,
            blogUrl = user.blogUrl,
            introduce = user.introduce,
            tags = user.tags
                .asSequence()
                .filter { it.tag.isActive }
                .map { it.tag.tag }
                .joinToString(" ") { tag -> "#$tag" },
            isSpeaker = isSpeaker,
            isManager = isManager,
            isFollowing = isFollowing,
            isBlock = blockMemberRepository.isBlocked(blockedMemberId = userResponse.id, memberId = memberResponse.id)
        )
    }

    fun getDonationTotal(roomId: Long): GetLiveRoomDonationTotalResponse {
        return GetLiveRoomDonationTotalResponse(
            totalDonationCan = repository.getDonationTotal(roomId = roomId) ?: 0
        )
    }

    fun setSpeaker(request: SetManagerOrSpeakerOrAudienceRequest) {
        val lock = getOrCreateLock(memberId = request.memberId)
        lock.write {
            val roomInfo = roomInfoRepository.findByIdOrNull(request.roomId)
                ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

            val member = memberRepository.findByIdOrNull(request.memberId)
                ?: throw SodaException("로그인 정보를 확인해 주세요.")

            if (roomInfo.speakerCount > 4) {
                throw SodaException("스피커 정원이 초과하였습니다.")
            }

            roomInfo.removeListener(member)
            roomInfo.removeManager(member)
            roomInfo.addSpeaker(member, cloudFrontHost)

            roomInfoRepository.save(roomInfo)
        }
    }

    fun setListener(request: SetManagerOrSpeakerOrAudienceRequest) {
        val lock = getOrCreateLock(memberId = request.memberId)
        lock.write {
            val roomInfo = roomInfoRepository.findByIdOrNull(request.roomId)
                ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

            val member = memberRepository.findByIdOrNull(request.memberId)
                ?: throw SodaException("로그인 정보를 확인해 주세요.")

            roomInfo.removeSpeaker(member)
            roomInfo.removeManager(member)
            roomInfo.addListener(member, cloudFrontHost)

            roomInfoRepository.save(roomInfo)
        }
    }

    fun setManager(request: SetManagerOrSpeakerOrAudienceRequest, member: Member) {
        val lock = getOrCreateLock(memberId = member.id!!)
        lock.write {
            val room = repository.getLiveRoom(request.roomId) ?: throw SodaException("잘못된 요청입니다.")
            if (room.member!!.id!! != member.id!!) {
                throw SodaException("권한이 없습니다.")
            }

            val user = memberRepository.findByIdOrNull(request.memberId) ?: throw SodaException("해당하는 유저가 없습니다.")
            val roomInfo = roomInfoRepository.findByIdOrNull(request.roomId)
                ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

            val roomAccountResponse = LiveRoomMember(member = user, cloudFrontHost)
            if (roomInfo.managerList.contains(roomAccountResponse)) {
                throw SodaException("이미 매니저 입니다.")
            }

            if (
                !roomInfo.speakerList.contains(roomAccountResponse) &&
                !roomInfo.listenerList.contains(roomAccountResponse)
            ) {
                throw SodaException("해당하는 유저가 없습니다.")
            }

            roomInfo.removeListener(user)
            roomInfo.removeSpeaker(user)
            roomInfo.addManager(user, cloudFrontHost)

            roomInfoRepository.save(roomInfo)
        }
    }

    @Transactional
    fun donation(request: LiveRoomDonationRequest, member: Member) {
        val room = repository.findByIdOrNull(request.roomId)
            ?: throw SodaException("해당하는 라이브가 없습니다.")

        val host = room.member ?: throw SodaException("잘못된 요청입니다.")

        if (host.role != MemberRole.CREATOR) {
            throw SodaException("비비드넥스트와 계약한\n크리에이터에게만 후원을 하실 수 있습니다.")
        }

        canPaymentService.spendCan(
            memberId = member.id!!,
            needCan = request.can,
            canUsage = CanUsage.DONATION,
            liveRoom = room,
            container = request.container
        )

        if (request.message.isNotBlank()) {
            val lock = getOrCreateLock(memberId = member.id!!)
            lock.write {
                val roomInfo = roomInfoRepository.findByIdOrNull(room.id!!)
                    ?: throw SodaException("해당하는 라이브의 정보가 없습니다.")

                roomInfo.addDonationMessage(
                    nickname = member.nickname,
                    can = request.can,
                    donationMessage = request.message
                )

                roomInfoRepository.save(roomInfo)
            }
        }
    }

    @Transactional
    fun refundDonation(roomId: Long, member: Member) {
        val donator = memberRepository.findByIdOrNull(member.id)
            ?: throw SodaException("후원에 실패한 캔이 환불되지 않았습니다\n고객센터로 문의해주세요.")

        val useCan = canRepository.getCanUsedForLiveRoomNotRefund(
            memberId = member.id!!,
            roomId = roomId,
            canUsage = CanUsage.DONATION
        ) ?: throw SodaException("후원에 실패한 캔이 환불되지 않았습니다\n고객센터로 문의해주세요.")
        useCan.isRefund = true

        val useCanCalculates = useCanCalculateRepository.findByUseCanIdAndStatus(useCan.id!!)
        useCanCalculates.forEach {
            it.status = UseCanCalculateStatus.REFUND
            val charge = Charge(0, it.can, status = ChargeStatus.REFUND_CHARGE)
            charge.title = "${it.can} 캔"
            charge.useCan = useCan

            when (it.paymentGateway) {
                PaymentGateway.GOOGLE_IAP -> donator.googleRewardCan += charge.rewardCan
                PaymentGateway.APPLE_IAP -> donator.appleRewardCan += charge.rewardCan
                else -> donator.pgRewardCan += charge.rewardCan
            }
            charge.member = donator

            val payment = Payment(
                status = PaymentStatus.COMPLETE,
                paymentGateway = it.paymentGateway
            )
            payment.method = "환불"
            charge.payment = payment

            chargeRepository.save(charge)
        }
    }

    fun getDonationStatus(roomId: Long, member: Member): GetLiveRoomDonationStatusResponse {
        val room = repository.getLiveRoom(roomId) ?: throw SodaException("잘못된 요청입니다.")
        val donationList = repository.getDonationList(roomId = room.id!!, cloudFrontHost = cloudFrontHost)
        val totalCan = donationList.sumOf { it.can }

        return GetLiveRoomDonationStatusResponse(
            donationList = donationList,
            totalCount = donationList.size,
            totalCan = totalCan
        )
    }

    @Transactional
    fun quitRoom(roomId: Long, member: Member) {
        val room = repository.getLiveRoom(roomId)
        val lock = getOrCreateLock(memberId = member.id!!)
        lock.write {
            val roomInfo = roomInfoRepository.findByIdOrNull(roomId)
            if (roomInfo != null) {
                if (room?.member != null && room.member!! == member) {
                    room.isActive = false
                    kickOutService.deleteKickOutData(roomId = room.id!!)
                    roomInfoRepository.deleteById(roomInfo.roomId)
                } else {
                    roomInfo.removeSpeaker(member)
                    roomInfo.removeListener(member)
                    roomInfo.removeManager(member)
                    roomInfoRepository.save(roomInfo)
                }
            }
        }
    }

    private fun getOrCreateLock(memberId: Long): ReentrantReadWriteLock {
        return tokenLocks.computeIfAbsent(memberId) { ReentrantReadWriteLock() }
    }
}
